// Declare the classes which handle triangulations and
// enumeration of degrees of freedom:
#include <deal.II/grid/tria.h>
#include <deal.II/dofs/dof_handler.h>
// functions to create grids:
#include <deal.II/grid/grid_generator.h>

// Classes which are needed for loops over all cells and to
// get the information from the cell objects.
#include <deal.II/grid/tria_accessor.h>
#include <deal.II/grid/tria_iterator.h>
#include <deal.II/dofs/dof_accessor.h>

// Contains the description of the Lagrange interpolation finite element.
#include <deal.II/fe/fe_q.h>

// Needed to create sparsity patterns of sparse matrices.
#include <deal.II/dofs/dof_tools.h>

// Needed for assembling the matrix using quadrature on each cell.
#include <deal.II/fe/fe_values.h>
#include <deal.II/base/quadrature_lib.h>
#include <deal.II/base/timer.h>

// Needed for the treatment of boundary values.
#include <deal.II/base/function.h>
#include <deal.II/numerics/vector_tools.h> //vectors.h>
#include <deal.II/numerics/matrix_tools.h> //matrices.h>

// Linear algebra related which we employ to solve the system of equations arising from the
// finite element discretization of the Laplace equation. We will use vectors and
// full matrices for assembling the system of equations locally on each cell, and
// transfer the results into a sparse matrix. We will then use a Conjugate
// Gradient solver to solve the problem, for which we need a preconditioner (in this
// program, we use the identity preconditioner which does nothing, but we
// need to include the file anyway).
#include <deal.II/lac/vector.h>
#include <deal.II/lac/full_matrix.h>
#include <deal.II/lac/sparse_matrix.h>
#include <deal.II/lac/compressed_sparsity_pattern.h>
#include <deal.II/lac/solver_cg.h>
#include <deal.II/lac/precondition.h>
#include <deal.II/lac/sparse_direct.h>

// Output to a file and to the console.
#include <deal.II/numerics/data_out.h>
#include <fstream>
#include <iostream>

#include <deal.II/lac/petsc_vector.h>
#include <deal.II/lac/petsc_parallel_vector.h>
#include <deal.II/lac/petsc_parallel_sparse_matrix.h>

#include <deal.II/lac/petsc_solver.h>
#include <deal.II/lac/petsc_precondition.h>

#include <deal.II/grid/grid_tools.h>
#include <deal.II/dofs/dof_renumbering.h>

using namespace dealii;

// Everything is encapsulated into a class for this program. The class consists of functions
// each performing certain aspects of a finite element program.

class laplace_problem
{
public:
    // Constructor
    laplace_problem ();

    // Runs the application.
    // Input is number of refinement levels.
    void run (int n_refs);
private:
    // Names are self explanatory
    void make_grid ();
    void setup_system ();
    void assemble_system ();
    void solve_cg ();
    void output_results () const;

    // Describing the triangulation.
    Triangulation<2>     triangulation;
    // Finite element object.
    FE_Q<2>              fe;
    // Handles degree of freedom.
    DoFHandler<2>        dof_handler;

    // Sparsity pattern and values of the system matrix resulting from
    // the discretization of the Laplace equation.
    PETScWrappers::MPI::SparseMatrix system_matrix;

    // Right hand side and solution vectors.
    PETScWrappers::MPI::Vector       solution;
    PETScWrappers::MPI::Vector        system_rhs;

    MPI_Comm mpi_communicator;

    const unsigned int n_mpi_processes;
    const unsigned int this_mpi_process;

    ConditionalOStream   pcout;
    TimerOutput computing_timer;

    unsigned int n_refinement_steps;
};

// Specify that bi-linear elements (denoted by the
// parameter to the finite element object, which indicates the polynomial degree) is used.
// Associate the dof_handler variable to the triangulation
laplace_problem::laplace_problem ()
    :
      pcout (std::cout, (Utilities::MPI::this_mpi_process(MPI_COMM_WORLD) == 0)),
      fe (1),
      dof_handler (triangulation),
      computing_timer (pcout, TimerOutput::summary, TimerOutput::wall_times),
      mpi_communicator (MPI_COMM_WORLD),
      n_mpi_processes (Utilities::MPI::n_mpi_processes(mpi_communicator)),
      this_mpi_process (Utilities::MPI::this_mpi_process(mpi_communicator))
{}

// Generate triangulation.
void laplace_problem::make_grid ()
{
    GridGenerator::hyper_cube (triangulation, -1, 1);
    triangulation.refine_global (n_refinement_steps);

    pcout << "Number of active cells: "
              << triangulation.n_active_cells()
              << std::endl;

    // Output the total number of cells.
    pcout << "Total number of cells: "
              << triangulation.n_cells()
              << std::endl;

    GridTools::partition_triangulation (n_mpi_processes, triangulation);
}

void laplace_problem::setup_system ()
{
    dof_handler.distribute_dofs (fe);
    DoFRenumbering::subdomain_wise (dof_handler);

    const unsigned int n_local_dofs
      = DoFTools::count_dofs_with_subdomain_association (dof_handler, this_mpi_process);

    pcout << "Number of degrees of freedom: " << dof_handler.n_dofs() << " (by partition:";
    for (unsigned int p=0; p<n_mpi_processes; ++p)
        pcout << (p==0 ? ' ' : '+') << (DoFTools::count_dofs_with_subdomain_association (dof_handler, p));
    pcout << ")" << std::endl;


    // Initialize system_matrix using the sparsity pattern.
    system_matrix.reinit (mpi_communicator,
                          dof_handler.n_dofs(),
                          dof_handler.n_dofs(),
                          n_local_dofs,
                          n_local_dofs,
                          dof_handler.max_couplings_between_dofs());

    // Set the sizes of the right hand side vector and
    // the solution vector.
    solution.reinit (mpi_communicator, dof_handler.n_dofs(), n_local_dofs);
    system_rhs.reinit (mpi_communicator, dof_handler.n_dofs(), n_local_dofs);
}


void laplace_problem::assemble_system ()
{
    QGauss<2>  quadrature_formula(2);

    FEValues<2> fe_values (fe, quadrature_formula, update_values | update_gradients | update_JxW_values);

    const unsigned int   dofs_per_cell = fe.dofs_per_cell;

    const unsigned int   n_q_points    = quadrature_formula.size();

    FullMatrix<double>   cell_matrix (dofs_per_cell, dofs_per_cell);
    Vector<double>       cell_rhs (dofs_per_cell);

    std::vector<unsigned int> local_dof_indices (dofs_per_cell);

    // Loop through all cells.
    DoFHandler<2>::active_cell_iterator
            cell = dof_handler.begin_active(),
            endc = dof_handler.end();
    for (; cell!=endc; ++cell) {
        if (cell->subdomain_id() == this_mpi_process){
        // Compute values for current cell.
        fe_values.reinit (cell);

        // Reset the local matrix and right hand side.
        cell_matrix = 0;
        cell_rhs = 0;

        for (unsigned int i=0; i<dofs_per_cell; ++i)
            for (unsigned int j=0; j<dofs_per_cell; ++j)
                for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
                    cell_matrix(i,j) += (fe_values.shape_grad (i, q_point) *
                                         fe_values.shape_grad (j, q_point) *
                                         fe_values.JxW (q_point));

        // Loop through degree of freedoms on each cell and compute cell_rhs using quadrature.
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
                cell_rhs(i) += (fe_values.shape_value (i, q_point) * 1 * fe_values.JxW (q_point));

        // Copy dof indecis of the cell to "local_dof_indices"
        cell->get_dof_indices (local_dof_indices);

        // add local cell_matrix to the system matrix
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            for (unsigned int j=0; j<dofs_per_cell; ++j)
                system_matrix.add (local_dof_indices[i], local_dof_indices[j], cell_matrix(i,j));

        // add local right hand side to system right hand side.
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            system_rhs(local_dof_indices[i]) += cell_rhs(i);
        }
    }

    system_matrix.compress(VectorOperation::add);
    system_rhs.compress(VectorOperation::add);

    // Map the degree of freedom of the boudaries to the values they should have.
    std::map<unsigned int,double> boundary_values;
    VectorTools::interpolate_boundary_values (dof_handler,
                                              0, // tag
                                              ZeroFunction<2>(),
                                              boundary_values);

    // Apply boundary values to the related matrix and vectors.
    MatrixTools::apply_boundary_values (boundary_values, system_matrix, solution, system_rhs);
}

// Conjugate gradiant solver.
void laplace_problem::solve_cg ()
{
    // Stopping criterion
    SolverControl           solver_control (system_rhs.size(), 1e-12);

    // Solver
    PETScWrappers::SolverCG cg (solver_control, mpi_communicator);

    PETScWrappers::PreconditionNone precond(system_matrix);
    // Solve system with out preconditioning.
    cg.solve (system_matrix, solution, system_rhs, precond);

    PETScWrappers::Vector localized_solution (solution);
    solution = localized_solution;

    // Number of iterations needed to reach solution.
    pcout<< "CG iterations without preconditioner:" << solver_control.last_step() << std::endl;
}

// Output results to file.
void laplace_problem::output_results () const
{
    const PETScWrappers::Vector localized_solution (solution);
    if (this_mpi_process == 0)
    {
        DataOut<2> data_out;
        data_out.attach_dof_handler (dof_handler);
        data_out.add_data_vector (localized_solution, "solution");
        std::vector<unsigned int> partition_int (triangulation.n_active_cells());
        GridTools::get_subdomain_association (triangulation, partition_int);
        const Vector<double> partitioning(partition_int.begin(), partition_int.end());
        data_out.add_data_vector (partitioning, "partitioning");

        data_out.build_patches ();
        std::ofstream output ("solution.gpl");
        data_out.write_gnuplot (output);
    }
}

// Run the application with number of refinements passed to this method.
void laplace_problem::run (int n_refs)
{
    double mesh_size;

    n_refinement_steps = n_refs;

    pcout<< "Number of refinements: " <<n_refinement_steps<< std::endl;

    mesh_size = 2*std::pow(0.5, double(n_refinement_steps));
    pcout << "Mesh size: " << mesh_size<< std::endl;

    make_grid ();
    setup_system();
    assemble_system ();

    computing_timer.enter_section("Solve system (CG)");
    solve_cg ();
    computing_timer.exit_section("Solve system (CG)");

    output_results ();
}

// Main method.
int main (int argc, char *argv[])
{
    Utilities::MPI::MPI_InitFinalize mpi_initialization (argc, argv);

    int n_refs;
    if (argc>=2)
        n_refs = atoi(argv[1]);
    else
        n_refs = 6;

    deallog.depth_console(0);
    laplace_problem laplace_problem;
    laplace_problem.run (n_refs);

    return 0;
}
