/* $Id: step-3.cc 24232 2011-09-02 09:47:37Z kronbichler $ */
/* Author: Wolfgang Bangerth, 1999, Guido Kanschat, 2011 */

/*    $Id: step-3.cc 24232 2011-09-02 09:47:37Z kronbichler $       */
/*                                                                */
/*    Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2006, 2007, 2008, 2010, 2011 by the deal.II authors */
/*                                                                */
/*    This file is subject to QPL and may not be  distributed     */
/*    without copyright and license information. Please refer     */
/*    to the file deal.II/doc/license.html for the  text  and     */
/*    further information on this license.                        */


// Ddeclare the classes which handle triangulations and
// enumeration of degrees of freedom:
#include <deal.II/grid/tria.h>
#include <deal.II/dofs/dof_handler.h>
// functions to create grids:
#include <deal.II/grid/grid_generator.h>

// Classes which are needed for loops over all cells and to
// get the information from the cell objects.
#include <deal.II/grid/tria_accessor.h>
#include <deal.II/grid/tria_iterator.h>
#include <deal.II/dofs/dof_accessor.h>

// Contains the description of the Lagrange interpolation finite element.
#include <deal.II/fe/fe_q.h>

// Needed to create sparsity patterns of sparse matrices.
#include <deal.II/dofs/dof_tools.h>

// Needed for assembling the matrix using quadrature on each cell.
#include <deal.II/fe/fe_values.h>
#include <deal.II/base/quadrature_lib.h>
#include <deal.II/base/timer.h>

// Needed for the treatment of boundary values.
#include <deal.II/base/function.h>
#include <deal.II/numerics/vector_tools.h> //vectors.h>
#include <deal.II/numerics/matrix_tools.h> //matrices.h>

// Linear algebra related which we employ to solve the system of equations arising from the
// finite element discretization of the Laplace equation. We will use vectors and
// full matrices for assembling the system of equations locally on each cell, and
// transfer the results into a sparse matrix. We will then use a Conjugate
// Gradient solver to solve the problem, for which we need a preconditioner (in this
// program, we use the identity preconditioner which does nothing, but we
// need to include the file anyway).
#include <deal.II/lac/vector.h>
#include <deal.II/lac/full_matrix.h>
#include <deal.II/lac/sparse_matrix.h>
#include <deal.II/lac/compressed_sparsity_pattern.h>
#include <deal.II/lac/solver_cg.h>
#include <deal.II/lac/precondition.h>
#include <deal.II/lac/sparse_direct.h>
#include <deal.II/lac/trilinos_precondition.h>

// Output to a file and to the console.
#include <deal.II/numerics/data_out.h>
#include <fstream>
#include <iostream>

using namespace dealii;

// Everything is encapsulated into a class for this program. The class consists of functions
// each performing certain aspects of a finite element program.

class laplace_problem
{
public:
    // Constructor
    laplace_problem ();

    // Runs the application.
    // Input is number of refinement levels.
    void run (int n_refs);
private:
    // Names are self explanatory
    void make_grid ();
    void setup_system ();
    void assemble_system ();
    void solve_direct ();
    void solve_cg ();
    void solve_amg ();
    void output_results () const;

    // Describing the triangulation.
    Triangulation<2>     triangulation;
    // Finite element object.
    FE_Q<2>              fe;
    // Handles degree of freedom.
    DoFHandler<2>        dof_handler;

    // Sparsity pattern and values of the system matrix resulting from
    // the discretization of the Laplace equation.
    SparsityPattern      sparsity_pattern;
    SparseMatrix<double> system_matrix;

    // Right hand side and solution vectors.
    Vector<double>       solution;
    Vector<double>       system_rhs;

    ConditionalOStream   pcout;
    TimerOutput computing_timer;

    unsigned int n_refinement_steps;

    std_cxx1x::shared_ptr<TrilinosWrappers::PreconditionAMG>    Amg_preconditioner;
};

// Specify that bi-linear elements (denoted by the
// parameter to the finite element object, which indicates the polynomial degree) is used.
// Associate the dof_handler variable to the triangulation
laplace_problem::laplace_problem ()
    :
      pcout (std::cout),
      fe (1),
      dof_handler (triangulation),
      computing_timer (pcout,
                       TimerOutput::summary,
                       TimerOutput::wall_times)
{}

// Generate triangulation.
void laplace_problem::make_grid ()
{
    // Create the grid and refine all cells five times. Since the
    // initial grid (which is the square [-1,1]x[-1,1]) consists
    // of only one cell, the final grid has 32 times 32 cells, for a
    // total of 1024.
    GridGenerator::hyper_cube (triangulation, -1, 1);
    triangulation.refine_global (n_refinement_steps);

    // Output the number of active cells.
    // Here, by active we mean the cells that aren't refined any further. There are actualy more
    // cells (the parent cells of the finest cells, their parents, etc,)
    std::cout << "Number of active cells: "
              << triangulation.n_active_cells()
              << std::endl;

    // Output the total number of cells.
    std::cout << "Total number of cells: "
              << triangulation.n_cells()
              << std::endl;
}

// We enumerate all the degrees of freedom and set up matrix and vector
// objects to hold the system data. Enumerating is done by using
// DoFHandler::distribute_dofs() method. Since we use the FE_Q class and have set the polynomial
// degree to 1(bilinear elements) in the constructor
void laplace_problem::setup_system ()
{
    dof_handler.distribute_dofs (fe);

    // Output number of degree of freedom.
    // There should be one DoF for each vertex. Since we have a 32 times
    // 32 grid, the number of DoFs should be 33 times 33, or 1089.
    std::cout << "Number of degrees of freedom: "
              << dof_handler.n_dofs()
              << std::endl;

    // Set up a sparsity pattern by:
    // Creating a temporary structure, tagging those entries that might be nonzero.
    CompressedSparsityPattern c_sparsity(dof_handler.n_dofs());
    DoFTools::make_sparsity_pattern (dof_handler, c_sparsity);

    // Then copy the data over to the SparsityPattern object.
    sparsity_pattern.copy_from(c_sparsity);

    // Initialize system_matrix using the sparsity pattern.
    system_matrix.reinit (sparsity_pattern);

    // Set the sizes of the right hand side vector and
    // the solution vector.
    solution.reinit (dof_handler.n_dofs());
    system_rhs.reinit (dof_handler.n_dofs());
}


void laplace_problem::assemble_system ()
{
    // Quadrature formula for the evaluation of the integrals on each cell.
    // Gauss formula with two quadrature points in each direction (total of four points since we are in 2D).
    QGauss<2>  quadrature_formula(2);

    // The class that exchange info between objects and retrieve needed values.
    // Initialize the object using the finite element object, the quadrature points
    // and their weights (jointly described by the Quadrature object).
    // FEValues also gets List of things to compute on each cell which is passed as flags.
    // The flags are:
    // "update_values" the values of the shape functions.
    // "update_gradients"
    // "update_JxW_values" The determinants of the Jacobians times the quadrature weights.
    FEValues<2> fe_values (fe, quadrature_formula, update_values | update_gradients | update_JxW_values);

    // names for these purposes:
    // The number of degrees of freedom on each cell.
    const unsigned int   dofs_per_cell = fe.dofs_per_cell;
    // The number of quadrature points.
    const unsigned int   n_q_points    = quadrature_formula.size();

    // To make it more efficient, create local matrices for each cell.
    FullMatrix<double>   cell_matrix (dofs_per_cell, dofs_per_cell);
    Vector<double>       cell_rhs (dofs_per_cell);

    // Temprory array to keep the local dof indecies in order to
    // transfer the data to global matrix later.
    std::vector<unsigned int> local_dof_indices (dofs_per_cell);

    // Loop through all cells.
    DoFHandler<2>::active_cell_iterator
            cell = dof_handler.begin_active(),
            endc = dof_handler.end();
    for (; cell!=endc; ++cell)
    {
        // Compute values for current cell.
        fe_values.reinit (cell);

        // Reset the local matrix and right hand side.
        cell_matrix = 0;
        cell_rhs = 0;

        // Loop through degree of freedoms on each cell and compute cell_matrix using quadrature.
        // get the gradient of shape function "i" at quadrature point "q_point" using
        // fe_values.shape_grad(i,q_point)
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            for (unsigned int j=0; j<dofs_per_cell; ++j)
                for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
                    cell_matrix(i,j) += (fe_values.shape_grad (i, q_point) *
                                         fe_values.shape_grad (j, q_point) *
                                         fe_values.JxW (q_point));

        // Loop through degree of freedoms on each cell and compute cell_rhs using quadrature.
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
                cell_rhs(i) += (fe_values.shape_value (i, q_point) * 1 *
                                fe_values.JxW (q_point));

        // Copy dof indecis of the cell to "local_dof_indices"
        cell->get_dof_indices (local_dof_indices);

        // add local cell_matrix to the system matrix
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            for (unsigned int j=0; j<dofs_per_cell; ++j)
                system_matrix.add (local_dof_indices[i],
                                   local_dof_indices[j],
                                   cell_matrix(i,j));

        // add local right hand side to system right hand side.
        for (unsigned int i=0; i<dofs_per_cell; ++i)
            system_rhs(local_dof_indices[i]) += cell_rhs(i);
    }


    // Map the degree of freedom of the boudaries to the values they should have.
    std::map<unsigned int,double> boundary_values;
    // Fill the value of boundaries which is taged "0" for the specified dofs to zero.
    // If you want to impose certain boundary values only on parts of the boundary, (for example,
    // you may have inflow and outflow boundaries in fluid dynamics) then you denote these different parts of
    // the boundary by different tag numbers and tell the interpolate_boundary_values
    // function to only compute the boundary values For the tag number which is passed to it.
    VectorTools::interpolate_boundary_values (dof_handler,
                                              0, // tag
                                              ZeroFunction<2>(),
                                              boundary_values);

    // Apply boundary values to the related matrix and vectors.
    MatrixTools::apply_boundary_values (boundary_values,
                                        system_matrix,
                                        solution,
                                        system_rhs);
}


// Solve the system directly using UMFPACK.
void laplace_problem::solve_direct ()
{
    SparseDirectUMFPACK direct_solver;
    // initialize using system_matrix
    direct_solver.initialize(system_matrix);
    // Multiply system_matrix by right hand side and put the result in solution variable.
    direct_solver.vmult (solution, system_rhs);
}

// Conjugate gradiant solver.
void laplace_problem::solve_cg ()
{
    // Stopping criterion
    SolverControl           solver_control (system_matrix.m(), 1e-12);

    // Solver
    SolverCG<>              solver (solver_control);

    // Solve system with out preconditioning.
    solver.solve (system_matrix, solution, system_rhs, PreconditionIdentity());

    // Number of iterations needed to reach solution.
    std::cout<< "CG iterations without preconditioner:"<<solver_control.last_step()<<std::endl;
}

// Algebraic multigrid solver.
void laplace_problem::solve_amg ()
{
    Amg_preconditioner.reset ();
    Amg_preconditioner = std_cxx1x::shared_ptr<TrilinosWrappers::PreconditionAMG>
            (new TrilinosWrappers::PreconditionAMG());

    std::vector<std::vector<bool> > constant_modes;
    std::vector<bool>  components (3,true);
    components[2] = false;
    DoFTools::extract_constant_modes (dof_handler, components, constant_modes);

    TrilinosWrappers::PreconditionAMG::AdditionalData Amg_data;
    Amg_data.constant_modes = constant_modes;
    Amg_data.elliptic = true;
    Amg_data.higher_order_elements = true;
    Amg_data.smoother_sweeps = 2;
    Amg_data.aggregation_threshold = 0.02;

    Amg_preconditioner->initialize(system_matrix, Amg_data);

    SolverControl           solver_control (system_matrix.m(), 1e-12);

    SolverCG<>              solver (solver_control);

    solver.solve (system_matrix, solution, system_rhs, *Amg_preconditioner);

    std::cout<< "CG iterations with AMG preconditioner:"<<solver_control.last_step()<<std::endl;
}

// Output results to file.
void laplace_problem::output_results () const
{
    // Output object for 2D data.
    DataOut<2> data_out;

    // Attache dof handler and the solution vector.
    data_out.attach_dof_handler (dof_handler);
    data_out.add_data_vector (solution, "solution");

    // Process the dofhandler and solution and create data for output.
    data_out.build_patches ();


    // Open a file stream and write the output data to this stream.
    std::ofstream output ("solution.gpl");
    data_out.write_gnuplot (output);
}

// Run the application with number of refinements passed to this method.
void laplace_problem::run (int n_refs)
{
    Vector<double>  init_sol;
    Vector<double>  init_rhs;
    double mesh_size;

    n_refinement_steps = n_refs;

    std::cout<< "Number of refinements: " <<n_refinement_steps<< std::endl;

    mesh_size = 2*std::pow(0.5, double(n_refinement_steps));
    pcout << "Mesh size: " << mesh_size<< std::endl;

    make_grid ();
    setup_system();
    assemble_system ();

    init_rhs = system_rhs;
    init_sol = solution;

    computing_timer.enter_section("Solve system directly");
    solve_direct ();
    computing_timer.exit_section("Solve system directly");

    solution = init_sol;
    system_rhs = init_rhs;
    computing_timer.enter_section("Solve system (CG)");
    solve_cg ();
    computing_timer.exit_section("Solve system (CG)");

    solution = init_sol;
    system_rhs = init_rhs;
    computing_timer.enter_section("Solve system (AMG)");
    solve_amg ();
    computing_timer.exit_section("Solve system (AMG)");

    output_results ();
}

// Main method.
int main (int argc, char *argv[])
{
    // Needed since the package is configured using mpi
    Utilities::MPI::MPI_InitFinalize mpi_initialization (argc, argv);

    int n_refs;
    if (argc>=2)
        n_refs = atoi(argv[1]);
    else
        n_refs = 6;

    deallog.depth_console(0);
    laplace_problem laplace_problem;
    laplace_problem.run (n_refs);

    return 0;
}
